# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of matrix multiply layers in flax with quantization.

Extends flax layers flax.linen.Dense.
"""

import contextlib
import dataclasses
import typing
from typing import Any, Callable, Iterable, Mapping, Optional, Sequence, Tuple, Type, Union

from absl import flags
from aqt.jax_legacy.jax import compute_cost_utils
from aqt.jax_legacy.jax import get_bounds
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax import quantization
from aqt.jax_legacy.jax import shape_utils
from aqt.jax_legacy.jax import stats_tag
from aqt.jax_legacy.jax import utils
from aqt.jax_legacy.jax.flax import struct as flax_struct
from aqt.jax_legacy.jax.quantization import QuantOps
from aqt.jax_legacy.jax.quantization import QuantType
from aqt.jax_legacy.jax.sparsity import SparseHParams
from aqt.jax_legacy.jax.sparsity import Sparsity
import flax
from flax import linen as nn
from flax.core import frozen_dict
from flax.linen import partitioning
import jax
from jax import lax
import jax.numpy as jnp
import numpy as np

FLAGS = flags.FLAGS

# Alias for initializer function.
# Should follow the template `def init(key, shape, dtype=dtype) -> ndarray:`.
# See flax.linen.initializers and jax.nn.initializers for more details.
InitializerType = Callable[[jnp.ndarray, Sequence[int], Type[Any]], jnp.ndarray]

default_kernel_init = nn.initializers.lecun_normal()

dataclass = flax_struct.dataclass if not typing.TYPE_CHECKING else dataclasses.dataclass

Array = jnp.ndarray
DType = jnp.dtype
PRNGKey = jnp.ndarray
Shape = Sequence[int]
Initializer = Callable[[PRNGKey, Shape, DType], Array]
PrecisionLike = Union[None, str, lax.Precision, Tuple[str, str],
                      Tuple[lax.Precision, lax.Precision]]


# Based on flax.linen.Dense
class DenseAqt(nn.Module):
  """A linear transformation with optional per-feature weight quantization.

  Attributes:
    features: the number of output features.
    hparams: hyperparameters
    update_bounds: Bool whether to update activation bounds.
    paxis_name: axis_name to which a user `pmaps` the parent module (model),
      refer to jax.pmap() for more documentation. This arg is used for
      get_bounds acts quantization (QuantOps.create_input_fake_quant)
    train: Whether model is training.
    collect_acts_stats: Whether to tag activations to record statistics.
    bias: whether to add a bias to the output (default: True).
    dtype: the dtype of the computation (default: float32).
    kernel_init: initializer function for the weight matrix. Should follow the
      template `def init(key, shape, dtype=dtype): -> array`. See
        flax.linen.initializers and jax.nn.initializers for more details.
    bias_init: initializer function for the bias. Should follow the template
      `def init(key, shape, dtype=dtype): -> array`. See flax.linen.initializers
        and jax.nn.initializers for more details.
    precision: numerical precision of the computation see `jax.lax.Precision`
      for details. Defaults to jax.lax.Precision.DEFAULT.
  """

  @dataclass
  class HParams:
    """Hyperparameter class to quantize Dense Layer."""
    # Target integer precision of weights in bits.
    # If None, no weight quantization will be applied.
    weight_prec: Union[None, int, QuantOps.FloatQuant]
    # half_shift flag for weights
    weight_half_shift: bool
    # QuantOps hyperparameter to quantize inputs. If None, no activation
    # quantization will be applied.
    quant_act: Optional[QuantOps.ActHParams]
    # Quantization strategy, one of `fake_quant` or `aqt`.
    quant_type: QuantType
    weight_quant_granularity: quant_config.QuantGranularity


    act_sparsity: Optional[SparseHParams] = None
    weight_sparsity: Optional[SparseHParams] = None

  hparams: HParams
  paxis_name: Optional[str]
  features: Union[int, Tuple[int, ...]]
  train: bool
  dynamic_context: quant_config.DynamicContext
  dtype: Any
  use_bias: bool = True
  kernel_init: InitializerType = default_kernel_init
  bias_init: InitializerType = nn.initializers.zeros
  precision: Optional[lax.Precision] = jax.lax.Precision.DEFAULT
  kernel_axis_names: Optional[Sequence[str]] = None
  possibly_use_quantized_vars: bool = False

  # TODO(shivaniagrawal): Changed the strategy to AQT if quant_type is aqt.

  @nn.compact
  def __call__(
      self,
      inputs: jnp.ndarray,
      *,
      padding_mask: Optional[jnp.ndarray],
  ) -> jnp.ndarray:
    """Applies a linear transformation to the inputs with optional quantization.

    If weight_prec is not None, scales and quantizes weights to signed int with
    weight_prec bits.

    Args:
      inputs: The nd-array to be transformed.
      padding_mask: boolean tensor of the same shape as 'inputs' specifying
        which values of 'inputs' to use as part of the bounds calculation.
        'True' indicates the corresponding value from 'inputs' should be used.
        If None, all values are used.

    Returns:
      The transformed input.
    """
    batch_size = inputs.shape[0]
    if padding_mask is not None:
      shape_utils.assert_shapes_equal(padding_mask.shape, (batch_size, 1))
    # TODO(wanglisa): Replace fake quant with AQT.

    if self.dynamic_context.collect_acts_stats:
      stats_tag.StatsTag(
          channel_axis=-1, name='inputs', update_stats=self.train)(
              inputs, mask=padding_mask)
    hparams = self.hparams

    if (hparams.weight_prec is not None and
        isinstance(hparams.weight_prec, int) and hparams.weight_prec > 8):
      raise NotImplementedError(
          'If you want to use more than 8bits for quantization, please revisit '
          'jax.lax.Precision.DEFAULT to determine whether it is still sufficient.'
      )

    if not isinstance(self.features, tuple):
      features = (self.features,)
    else:
      features = self.features
    kernel_shape = (inputs.shape[-1],) + features
    if self.kernel_axis_names is None:
      kernel_axis_names = ['unmodeled'] * len(kernel_shape)
    else:
      kernel_axis_names = self.kernel_axis_names
      if len(kernel_axis_names) != len(kernel_shape):
        raise ValueError(f"Kernel axis names {kernel_axis_names} doesn't match "
                         f'kernel shape {kernel_shape}.')

    inputs = jnp.asarray(inputs, self.dtype)

    # Here are 3 possible conditions:
    #   1. not quantization: init normal kernel, quant_w = None
    #   2. quantization and train mode: init normal kernel, quantized kernel
    #      with bfloat16 to avoid gradient on integer and quant_w = None.
    #   3. quantization and inference mode: no normal kernel, init quantized
    #      kernel with int8 and quant_w = (qkernel, qscale).
    quant_w = None
    if (not self.possibly_use_quantized_vars) or (
        self.possibly_use_quantized_vars and self.train):
      kernel = partitioning.param_with_axes(
          'kernel',
          self.kernel_init,
          kernel_shape,
          axes=tuple(kernel_axis_names))
      kernel = jnp.asarray(kernel, self.dtype)

      kernel = Sparsity(
          sparsity_hparams=hparams.weight_sparsity, name='weight_sparsity')(
              kernel,
              update_mask=self.dynamic_context.update_weight_sparsity,
              apply_mask=self.dynamic_context.apply_sparsity,
              num_update_sparsity=self.dynamic_context.num_update_sparsity)

    if self.possibly_use_quantized_vars:
      qkernel_dtype = self.dtype
      qkernel_initializer = nn.initializers.zeros
      if not self.train:
        qkernel_dtype = jax.numpy.int8
        qkernel_initializer = nn.initializers.zeros

      qkernel = partitioning.param_with_axes(
          'qkernel',
          qkernel_initializer,
          kernel_shape,
          qkernel_dtype,
          axes=tuple(kernel_axis_names))
      qkernel = jnp.asarray(qkernel, qkernel_dtype)
      # Initialization of scale does not matter so we are initializing with
      # bias_init. And here only support quantizing on first dimension.
      scale_axis_names = list(kernel_axis_names)
      scale_axis_names[0] += '_qscale'
      qscale = partitioning.param_with_axes(
          'qscale',
          self.bias_init, (1,) + features,
          axes=tuple(scale_axis_names))
      qscale = jnp.asarray(qscale, self.dtype)
      if not self.train:
        quant_w = quantization.QuantW(qkernel, qscale)
        kernel = None

    inputs = Sparsity(
        sparsity_hparams=hparams.act_sparsity, name='act_sparsity')(
            inputs,
            update_mask=self.dynamic_context.update_act_sparsity,
            apply_mask=self.dynamic_context.apply_sparsity,
            num_update_sparsity=self.dynamic_context.num_update_sparsity)

    weight_quant_granularity = hparams.weight_quant_granularity
    # kernel.shape = (channels_in, channels_out)
    if weight_quant_granularity == quant_config.QuantGranularity.PER_CHANNEL:
      # Compute scale factors by reducing over the rows of the weight matrix,
      # resulting in one scale factor per column. This results in one scale
      # factor per output channel.
      expected_scale_shape = (1,) + features
      weight_quant_axis = (0,)
    elif weight_quant_granularity == quant_config.QuantGranularity.PER_TENSOR:
      # Compute a single scale factor for the entire weight matrix.
      expected_scale_shape = (1,) * len(kernel_shape)
      weight_quant_axis = None
    else:
      raise ValueError(
          f'Invalid quantization granularity {weight_quant_granularity}.')

    weight_params = QuantOps.WeightParams(
        prec=hparams.weight_prec,
        half_shift=hparams.weight_half_shift,
        axis=weight_quant_axis,
        expected_scale_shape=expected_scale_shape)

    # TODO(wanglisa): add option to control when scale is being recomputed

    bounds_params = None
    if hparams.quant_act is not None:
      if isinstance(hparams.quant_act.bounds, get_bounds.DynamicBounds.Hyper):
        bounds_params = get_bounds.DynamicBounds.Params(quant_axis=None)
      elif isinstance(hparams.quant_act.bounds, get_bounds.GetBounds.Hyper):
        bounds_params = get_bounds.GetBounds.Params(
            update_bounds=self.dynamic_context.update_bounds,
            update_stats=self.train,
            paxis_name=self.paxis_name,
            mask=padding_mask)

    contracting_dims = ((inputs.ndim - 1,), (0,))
    # `((lhs_contracting_dims, rhs_contracting_dims),
    batch_dims = ((), ())  # (lhs_batch_dims, rhs_batch_dims))`

    if hparams.quant_act is not None and isinstance(
        hparams.quant_act.bounds, get_bounds.DynamicBounds.Hyper):
      y = quantization.flaxformer_dot_general(
          act=inputs,
          w=kernel,
          dimension_numbers=(contracting_dims, batch_dims),
          weight_params=weight_params,
          act_hparams=hparams.quant_act,
          bounds_params=bounds_params,
          dot_precision=self.precision,
          quant_w=quant_w)

    else:
      y = quantization.quantized_dot_general(
          act=inputs,
          w=kernel,
          quant_type=hparams.quant_type,
          weight_params=weight_params,
          act_hparams=hparams.quant_act,
          bounds_params=bounds_params,
          dimension_numbers=(contracting_dims, batch_dims),
          dot_precision=self.precision,
          prefer_int8_to_int32_dot=self.dynamic_context
          .prefer_int8_to_int32_dot,
          quant_w=quant_w)

    # bias
    if self.use_bias:
      bias = partitioning.param_with_axes(
          'bias',
          self.bias_init, (np.prod(features),),
          axes=(kernel_axis_names[-1],))
      bias = jnp.asarray(bias, self.dtype)
      bias = jnp.reshape(bias, features)
      # (batch_size, features)
      y = y + bias[jnp.newaxis, :]
    return y


def _canonicalize_tuple(x):
  if isinstance(x, Iterable):
    return tuple(x)
  else:
    return (x,)


# The Flaxformer sharding API emits some names that are too detailed, so we
# remap them here. Any values that don't match keys here are unchanged.
_RESHAPED_KERNEL_AXIS_NAME_MAP = frozen_dict.freeze({
    'heads * kv': 'joined_kv',
})


def _normalize_axes(axes: Iterable[int], ndim: int) -> Tuple[int, ...]:
  # A tuple by convention. len(axes_tuple) then also gives the rank efficiently.
  return tuple([ax if ax >= 0 else ndim + ax for ax in axes])


# Based on flaxformer DenseGeneral
class DenseGeneralAqt(nn.Module):
  """A linear transformation with flexible axes.

    Kernel stored as 2d parameter for compatibility with Adafactor optimizer.

    Attributes:
      features: tuple with numbers of output features.
      use_bias: whether to add a bias to the output (default: False).
      axis: tuple with axes to apply the transformation on.
      dtype: the dtype of the computation (default: float32).
      kernel_init: initializer function for the weight matrix.
      bias_init: initializer function for the bias.
      precision: numerical precision of the computation see `jax.lax.Precision`
        for details.
      kernel_axis_names: logical axis names to use for kernel sharding. Each
        should be one of _VALID_AXIS_NAMES in sharding.py.
      reshaped_kernel_axis_name_map: Rules for renaming fused kernel axes. We
        keep this as a separate parameter than kernel_axis_names so that
        experiments can toggle `reshape_kernel` without having to keep
        `kernel_axis_names` in sync.
      reshape_kernel: whether to reshape the kernel parameter to 2D for
        Adafactor.
  """

  @dataclass
  class HParams:
    """Hyperparameter class to quantize Dense Layer."""
    # Target integer precision of weights in bits.
    # If None, no weight quantization will be applied.
    weight_prec: Union[None, int, QuantOps.FloatQuant]
    # half_shift flag for weights
    weight_half_shift: bool
    # QuantOps hyperparameter to quantize inputs. If None, no activation
    # quantization will be applied.
    quant_act: Optional[QuantOps.ActHParams]
    weight_quant_granularity: quant_config.QuantGranularity

  hparams: HParams
  train: bool
  features: Union[Iterable[int], int]
  use_bias: bool
  axis: Union[Iterable[int], int] = -1
  dtype: DType = jnp.float32
  kernel_init: Initializer = default_kernel_init
  bias_init: Initializer = nn.initializers.zeros
  precision: Any = None
  kernel_axis_names: Optional[Sequence[str]] = None
  reshaped_kernel_axis_name_map: Mapping[str, str] = (
      _RESHAPED_KERNEL_AXIS_NAME_MAP)
  reshape_kernel: bool = True
  possibly_use_quantized_vars: bool = False

  @nn.compact
  def __call__(self, inputs: Array) -> Array:
    """Applies a linear transformation to the inputs along multiple dimensions.

    Args:
      inputs: The nd-array to be transformed.

    Returns:
      The transformed input.
    """
    features = _canonicalize_tuple(self.features)
    axis = _canonicalize_tuple(self.axis)

    inputs = jnp.asarray(inputs, self.dtype)
    axis = _normalize_axes(axis, inputs.ndim)

    hparams = self.hparams

    if (hparams.weight_prec is not None and
        isinstance(hparams.weight_prec, int) and hparams.weight_prec > 8):
      raise NotImplementedError(
          'If you want to use more than 8bits for quantization, please revisit '
          'jax.lax.Precision.DEFAULT to determine whether it is still sufficient.'
      )

    kernel_shape = tuple([inputs.shape[ax] for ax in axis]) + features
    scale_shape = (1,) * len(axis) + features
    if self.reshape_kernel:
      kernel_param_shape = (np.prod([inputs.shape[ax] for ax in axis]),
                            np.prod(features))
      scale_param_shape = (1, np.prod(features))
    else:
      kernel_param_shape = kernel_shape
      scale_param_shape = scale_shape

    # Determine axes names metadata for partitioning/adafactor rules.
    if self.kernel_axis_names is None:
      kernel_axis_names = ['unmodeled'] * len(kernel_param_shape)
    else:
      kernel_axis_names = self.kernel_axis_names
      if len(kernel_axis_names) != len(kernel_shape):
        raise ValueError(f"Kernel axis names {kernel_axis_names} doesn't match "
                         f'kernel shape {kernel_shape}.')
      if self.reshape_kernel:

        def _reshaped_axis_names(names):
          result = ' * '.join(names)
          return self.reshaped_kernel_axis_name_map.get(result, result)

        kernel_axis_names = (
            _reshaped_axis_names(kernel_axis_names[:len(axis)]),
            _reshaped_axis_names(kernel_axis_names[len(axis):]),
        )

    # Here are 3 possible conditions:
    #   1. not quantization: init normal kernel, quant_w = None
    #   2. quantization and train mode: init normal kernel, quantized kernel
    #      with bfloat16 to avoid gradient on integer and quant_w = None.
    #   3. quantization and inference mode: no normal kernel, init quantized
    #      kernel with int8 and quant_w = (qkernenl, qscale).
    quant_w = None
    if (not self.possibly_use_quantized_vars) or (
        self.possibly_use_quantized_vars and self.train):
      kernel = partitioning.param_with_axes(
          'kernel',
          self.kernel_init,
          kernel_param_shape,
          jnp.float32,
          axes=tuple(kernel_axis_names))
      kernel = jnp.asarray(kernel, self.dtype)
      kernel = jnp.reshape(kernel, kernel_shape)

    if self.possibly_use_quantized_vars:
      qkernel_dtype = self.dtype
      qkernel_initializer = nn.initializers.zeros
      if not self.train:
        qkernel_dtype = jax.numpy.int8
        qkernel_initializer = nn.initializers.zeros
      qkernel = partitioning.param_with_axes(
          'qkernel',
          qkernel_initializer,
          kernel_param_shape,
          qkernel_dtype,
          axes=tuple(kernel_axis_names))
      qkernel = jnp.asarray(qkernel, qkernel_dtype)
      qkernel = jnp.reshape(qkernel, kernel_shape)

      # Initialization of scale does not matter so we are initializing with
      # bias_init.
      scale_axis_names = list(kernel_axis_names)
      len_contracted_dims = 1 if self.reshape_kernel else len(axis)
      for ax in range(0, len_contracted_dims):
        scale_axis_names[ax] += '_qscale'
      qscale = partitioning.param_with_axes(
          'qscale',
          self.bias_init,
          scale_param_shape,
          axes=tuple(scale_axis_names))
      qscale = jnp.asarray(qscale, self.dtype)
      qscale = jnp.reshape(qscale, scale_shape)
      if not self.train:
        quant_w = quantization.QuantW(qkernel, qscale)
        kernel = None

    contract_ind = tuple(range(0, len(axis)))

    weight_quant_granularity = hparams.weight_quant_granularity

    act_quant_axis = None
    if hparams.quant_act:
      if isinstance(hparams.quant_act.bounds, get_bounds.DynamicBounds.Hyper):
        act_quant_granularity = hparams.quant_act.bounds.granularity  # pytype: disable=attribute-error  # jax-ndarray
        if act_quant_granularity == quant_config.QuantGranularity.PER_CHANNEL:
          act_quant_axis = tuple(axis)
        elif act_quant_granularity == quant_config.QuantGranularity.PER_TENSOR:
          act_quant_axis = None
        else:
          raise ValueError(
              f'Invalid quantization granularity {weight_quant_granularity}.')
      elif isinstance(hparams.quant_act.bounds, get_bounds.GetBounds.Hyper):
        raise NotImplementedError(
            'We do not support get_bounds for dot general.')

    if weight_quant_granularity == quant_config.QuantGranularity.PER_CHANNEL:
      # Compute scale factors by reducing over the rows of the weight matrix,
      # resulting in one scale factor per column. This results in one scale
      # factor per output channel.
      expected_scale_shape = (1,) * len(axis) + features
      weight_quant_axis = contract_ind
    elif weight_quant_granularity == quant_config.QuantGranularity.PER_TENSOR:
      # Compute a single scale factor for the entire weight matrix.
      expected_scale_shape = (1,) * len(kernel_shape)
      weight_quant_axis = None
    else:
      raise ValueError(
          f'Invalid quantization granularity {weight_quant_granularity}.')

    bounds_params = get_bounds.DynamicBounds.Params(quant_axis=act_quant_axis)

    weight_params = QuantOps.WeightParams(
        prec=hparams.weight_prec,
        half_shift=hparams.weight_half_shift,
        axis=weight_quant_axis,
        expected_scale_shape=expected_scale_shape)

    out = quantization.flaxformer_dot_general(
        act=inputs,
        w=kernel,
        dimension_numbers=((axis, contract_ind), ((), ())),
        weight_params=weight_params,
        act_hparams=hparams.quant_act,
        bounds_params=bounds_params,
        dot_precision=self.precision,
        quant_w=quant_w)

    if self.use_bias:
      bias = partitioning.param_with_axes(
          'bias',
          self.bias_init, (np.prod(features),),
          jnp.float32,
          axes=(kernel_axis_names[-1],))
      bias = jnp.asarray(bias, self.dtype)
      bias = jnp.reshape(bias, features)
      # Reshape bias for broadcast.
      expand_dims = sorted(set(range(inputs.ndim)) - set(axis))
      for ax in expand_dims:
        bias = jnp.expand_dims(bias, ax)
      out = out + bias
    return out


# Based on flax.linen.Conv
class ConvAqt(nn.Module):
  """Convolution Module with optional quantization.

  Attributes:
    features: number of convolution filters.
    hparams: hyperparameters
    update_bounds: Bool whether to update activation bounds.
    paxis_name: axis_name to which a user `pmaps` the parent module (model),
      refer to jax.pmap() for more documentation. This arg is used for
      get_bounds acts quantization (QuantOps.create_input_fake_quant)
    train: Whether model is training.
    kernel_size: shape of the convolutional kernel.
    strides: a sequence of `n` integers, representing the inter-window strides.
    padding: either the string `'SAME'`, the string `'VALID'`, or a sequence of
      `n` `(low, high)` integer pairs that give the padding to apply before and
      after each spatial dimension.
    input_dilation: `None`, or a sequence of `n` integers, giving the dilation
      factor to apply in each spatial dimension of `inputs`. Convolution with
      input dilation `d` is equivalent to transposed convolution with stride
      `d`.
    kernel_dilation: `None`, or a sequence of `n` integers, giving the dilation
      factor to apply in each spatial dimension of the convolution kernel.
      Convolution with kernel dilation is also known as 'atrous convolution'.
    feature_group_count: integer, default 1. If specified divides the input
      features into groups.
    bias: whether to add a bias to the output (default: True).
    dtype: the dtype of the computation (default: float32).
    kernel_init: initializer for the convolutional kernel.
    bias_init: initializer for the bias.
  """

  @dataclass
  class HParams:
    """Hyperparameter class to quantize Conv Layer."""
    # Target integer precision of weights in bits.
    # If None, no weight quantization will be applied.
    # TODO(shivaniagrawal): we could replace this with quant_weight instead.
    weight_prec: Union[None, int, QuantOps.FloatQuant]
    # half_shift flag for weights
    weight_half_shift: bool
    # QuantOps hyperparameter to quantize inputs. If None, no activation
    # quantization will be applied.
    quant_act: Optional[QuantOps.ActHParams]
    # Quantization strategy, one of `fake_quant` or `aqt`.
    quant_type: QuantType

  hparams: HParams
  features: int
  kernel_size: Tuple[int, ...]
  dynamic_context: quant_config.DynamicContext
  train: bool
  paxis_name: Optional[str]
  dtype: Any
  strides: Optional[Tuple[int, ...]] = None
  padding: Union[str, Sequence[Tuple[int, ...]]] = 'SAME'
  input_dilation: Optional[Tuple[int, ...]] = None
  kernel_dilation: Optional[Tuple[int, ...]] = None
  feature_group_count: int = 1
  use_bias: bool = True
  kernel_init: InitializerType = default_kernel_init
  bias_init: InitializerType = flax.linen.initializers.zeros

  @nn.compact
  def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
    """Applies a convolution to the inputs with optional quantization.

    Args:
      inputs: input data with dimensions (batch, spatial_dims..., features).

    Returns:
      The convolved data.
    """
    hparams = self.hparams
    if hparams.weight_prec is not None and hparams.weight_prec > 8:
      raise NotImplementedError(
          'If you want to use more than 8bits for quantization, please revisit '
          'jax.lax.Precision.DEFAULT to determine whether it is still sufficient.'
      )
    jax_precision = jax.lax.Precision.DEFAULT

    if self.strides is None:
      strides = (1,) * (inputs.ndim - 2)
    else:
      strides = self.strides

    in_features = inputs.shape[-1]
    assert in_features % self.feature_group_count == 0
    kernel_shape = self.kernel_size + (in_features // self.feature_group_count,
                                       self.features)
    kernel = self.param('kernel', self.kernel_init, kernel_shape)

    inputs = jnp.asarray(inputs, self.dtype)
    kernel = jnp.asarray(kernel, self.dtype)

    # Activation quantization
    if hparams.quant_act is not None:
      inputs = QuantOps.create_inputs_fake_quant(
          inputs=inputs,
          hparams=hparams.quant_act,
          bounds_params=get_bounds.GetBounds.Params(
              update_bounds=self.dynamic_context.update_bounds,
              update_stats=self.train,
              paxis_name=self.paxis_name))

    # Weight quantization
    if hparams.weight_prec is not None:
      kernel_reduction_axis = tuple(range(kernel.ndim - 1))
      expected_scale_shape = (1,) * (kernel.ndim - 1) + (self.features,)
      assert hparams.quant_type == QuantType.FAKE_QUANT, (
          'we only support fake_quant style of aqt for ConvAqt.')
      quantized_type = hparams.quant_type.to_jax_type()
      kernel = QuantOps.create_weights_fake_quant(
          kernel,
          weight_params=QuantOps.WeightParams(
              prec=hparams.weight_prec,
              half_shift=hparams.weight_half_shift,
              axis=kernel_reduction_axis,
              expected_scale_shape=expected_scale_shape),
          quantized_type=quantized_type,
          quantize_weights=self.dynamic_context.quantize_weights)

    # Convolution
    dimension_numbers = flax.linen.linear._conv_dimension_numbers(
        inputs.shape)  # pylint: disable=protected-access
    metadata_context = contextlib.suppress()
    # Use metadata context to annotate op metadata with quantization info
    act_prec = None if hparams.quant_act is None else hparams.quant_act.prec

    if flags.FLAGS.metadata_enabled:
      metadata_context = compute_cost_utils.ConvMetadataMonkeyPatch(
          weight_prec=hparams.weight_prec, act_prec=act_prec)
    with metadata_context:
      y = lax.conv_general_dilated(
          inputs,
          kernel,
          strides,
          self.padding,
          lhs_dilation=self.input_dilation,
          rhs_dilation=self.kernel_dilation,
          dimension_numbers=dimension_numbers,
          feature_group_count=self.feature_group_count,
          precision=jax_precision)
    # TODO(shivaniagrawal): create quantized conv general dilated.

    # bias
    if self.use_bias:
      bias = self.param('bias', self.bias_init, (self.features,))
      bias = jnp.asarray(bias, self.dtype)
      # The inputs can have an arbitrary number of spatial dims, so we broadcast
      # the bias to match: (batch_size, spatial_dim,... features)
      # TODO(shivaniagrawal): Consider making ConvAqt rank static (e.g. 2D)
      # or maybe add error checking (e.g. expect inputs to have rank N, but this
      # may already be checked by lax.conv_general_dialated).
      bias = utils.broadcast_rank(bias, inputs)
      y = y + bias
    return y


# From flax.linen.Embed
default_embed_init = nn.initializers.variance_scaling(
    1.0, 'fan_in', 'normal', out_axis=0)


# This is based on flax.linen.Embed
# (https://github.com/google/flax/blob/65061e6128f6695eed441acf2bfffc3b1badd318/flax/nn/linear.py#L360)
class EmbedAqt(nn.Module):
  """Quantized Embedding Module.

  A parameterized function from integers [0, n) to d-dimensional vectors.

  Attributes:
    num_embeddings: number of embeddings.
    features: Number of feature dimensions for each embedding.
    hparams: hyperparameters
    dtype: dtype to use for embedding.
    paxis_name: axis_name to which a user `pmaps` the parent module (model),
      refer to jax.pmap() for more documentation. This arg is used for
      get_bounds acts quantization (QuantOps.create_input_fake_quant)
    train: Whether model is training.
    update_bounds: Bool whether to update activation bounds.
    embedding_init: embedding initializer
  """

  @dataclass
  class HParams:  # pylint: disable=missing-docstring
    # Target integer precision of weights in bits.
    # If None, no quantization will be applied.
    weight_prec: Union[None, int, QuantOps.FloatQuant]
    # half_shift flag for weights
    weight_half_shift: bool
    # QuantOps hyperparameter to quantize inputs for logits. If None, no
    # activation quantization will be applied.
    quant_act: Optional[QuantOps.ActHParams]
    # Quantization strategy, one of `fake_quant` or `aqt`.
    quant_type: QuantType

  num_embeddings: int
  features: int
  hparams: HParams
  dtype: Any
  paxis_name: Optional[str]
  train: bool
  dynamic_context: quant_config.DynamicContext
  embedding_init: InitializerType = default_embed_init

  def setup(self):
    self.embedding = self.param(
        'embedding',
        self.embedding_init,  # pylint: disable=missing-from-attributes
        (self.num_embeddings, self.features))
    hparams = self.hparams
    if hparams.quant_act is not None and isinstance(hparams.quant_act.bounds,
                                                    get_bounds.GetBounds.Hyper):
      self.get_bounds_logits = get_bounds.GetBounds(  # pylint: disable=missing-from-attributes
          hyper=self.hparams.quant_act.bounds)
    self.quantized_dot = quantization.QuantizedDot(  # pylint: disable=missing-from-attributes
        act_hparams=hparams.quant_act,
        quant_type=hparams.quant_type,
        dot_precision=None,
        prefer_int8_to_int32_dot=self.dynamic_context.prefer_int8_to_int32_dot,
        weight_params=QuantOps.WeightParams(
            prec=hparams.weight_prec,
            axis=(0,),
            expected_scale_shape=(1, self.embedding.shape[0]),
            half_shift=hparams.weight_half_shift,
            ))

  def __call__(
      self,
      inputs: jnp.ndarray,
  ) -> jnp.ndarray:
    """Embeds the inputs along the last dimension.

    Args:
      inputs: input data, all dimensions are considered batch dimensions.

    Returns:
      Output which is embedded input data.  The output shape follows the input,
      with an additional `features` dimension appended.
    """
    batch_size, sequence_length = inputs.shape
    if inputs.dtype not in [jnp.int32, jnp.int64, jnp.uint32, jnp.uint64]:
      raise ValueError('Input type must be an integer or unsigned integer.')
    embedding = self.embedding

    embedding = jnp.asarray(embedding, self.dtype)

    hparams = self.hparams
    # Initialize state for stats and bounds, this would be required for logits
    # in the following method attend.
    if hparams.quant_act is not None and isinstance(hparams.quant_act.bounds,
                                                    get_bounds.GetBounds.Hyper):
      self.get_bounds_logits(
          inputs,
          bounds_params=get_bounds.GetBounds.Params(
              update_stats=False, update_bounds=False, paxis_name=None),
      )

    weight_prec = hparams.weight_prec
    weight_half_shift = hparams.weight_half_shift
    if weight_prec is not None:
      quantized_type = hparams.quant_type.to_jax_type()
      # In contrast to all other scale factor calculations in this module, we
      # compute per-row instead of per-column (ie, per-output-channel) scale
      # factors here. This is because the embedding matrix might be shared with
      # the output (logit) layer of the transformer, in which case the
      # *transpose* of the embedding matrix will be used as the weight matrix in
      # a mamtul. The per-row scale factors used here would thus correspond to
      # using per-column (because of the transpose) scale factors used by the
      # weight matrix in the logits layer, which is what we need for AQT.
      embedding_quant_ops = QuantOps.create_weights_ops(
          embedding,
          weight_params=QuantOps.WeightParams(
              prec=weight_prec,
              axis=(1,),
              half_shift=weight_half_shift,
              ))
      embedding_quant_ops.assert_scale_shape_is(shape=(self.num_embeddings, 1))

      quantized_embedding = embedding_quant_ops.to_quantized(
          embedding, dtype=quantized_type)
      quantized_embedded_inputs = quantized_embedding[inputs]
      # Since the embedding matrix 'quantized_embedding' is gathered based on
      # 'inputs' to produce the embedding tensor, we apply the same gathering to
      # the per-row scale factors of the embedding matrix so the scale factors
      # will broadcast appropriately in the subsequent call to 'to_quantized'.
      # TODO(malmaud): As part of quantization.py refactor, change
      # 'get_scale_for_aqt' to cleanly support this and hence avoid the need to
      # directly access a protected member of QuantOps.
      scale = embedding_quant_ops._scale[inputs]  # pylint: disable=protected-access
      shape_utils.assert_shapes_equal(scale.shape,
                                      (batch_size, sequence_length, 1))
      shape_utils.assert_shapes_equal(
          quantized_embedded_inputs.shape,
          (batch_size, sequence_length, self.features))
      embedded_inputs = (quantized_embedded_inputs / scale).astype(self.dtype)
    else:
      embedded_inputs = embedding[inputs]
    shape_utils.assert_shapes_equal(
        embedded_inputs.shape, (batch_size, sequence_length, self.features))
    return embedded_inputs

  def attend(self, query: jnp.ndarray, padding_mask: Optional[jnp.ndarray],
             **unused_kwargs) -> jnp.ndarray:
    """Attend over the embedding using a query array.

    Args:
      query: array with last dimension equal the feature depth `features` of the
        embedding.
      padding_mask: boolean mask indicating which elements of 'query' are
        padding. Used for calculating activation statistics for the dynamic
        bounds quantization algorithm.
      **unused_kwargs: unused arguments passed from the apply method.

    Returns:
      An array with final dim `num_embeddings` corresponding to the batched
      inner-product of the array of query vectors against each embedding.
      Commonly used for weight-sharing between embeddings and logit transform
      in NLP models.
    """
    del unused_kwargs

    batch_size = query.shape[0]

    if padding_mask is not None:
      shape_utils.assert_shapes_equal(padding_mask.shape, (batch_size, 1))

    embedding = self.embedding
    embedding = jnp.asarray(embedding, self.dtype)

    # TODO(malmaud): Remove the 'mask' field from this struct so we can
    # make this struct a hyperparameter of the EncoderAqt class.
    bounds_params = get_bounds.GetBounds.Params(
        update_bounds=self.dynamic_context.update_bounds,
        update_stats=self.train,
        paxis_name=self.paxis_name,
        mask=padding_mask,
        module_name='logits')

    out = self.quantized_dot(
        act=query,
        w=jnp.transpose(embedding),
        bounds_params=bounds_params)

    return out


# Forked from Flax LayerNorm module.
class LayerNormAqt(nn.Module):
  """Layer normalization (https://arxiv.org/abs/1607.06450).

  Operates on the last axis of the input data.

  Adds quantization support to the Flax LayerNorm layer.

  Attributes:
    epsilon: A small float added to variance to avoid dividing by zero.
    dtype: the dtype of the computation (default: float32). Can also be eg
      bfloat16. Note this is the real Jax type that intermediate activations
      will be stored in and is separate from the quantized type specified in
      'hparams' which we are simulating via downcasting operations.
    bias:  If True, bias (beta) is added.
    scale: If True, multiply by scale (gamma). When the next layer is linear
      (also e.g. nn.relu), this can be disabled since the scaling will be done
      by the next layer.
    bias_init: Initializer for bias, by default, zero.
    scale_init: Initializer for scale, by default, one.
    hparams: An instance of LayerNormAqt.HParams with quantization-related
      hyperparameters.
  """

  @dataclass
  class QuantHParams:
    # TODO(malmaud): Generalize this to other quantization formats.
    prec: QuantOps.FloatQuant.FloatPrec
    reduction_prec: Optional[QuantOps.FloatQuant.FloatPrec]

  @dataclass
  class HParams:
    # We have to refer to the type name with quotes here since we can't directly
    # refer to types being defined in the same class.
    quant_hparams: Optional['LayerNormAqt.QuantHParams']

  hparams: HParams
  dtype: Any
  dynamic_context: quant_config.DynamicContext
  epsilon: float = 1e-6
  use_bias: bool = True
  use_scale: bool = True
  bias_init: InitializerType = nn.initializers.zeros
  scale_init: InitializerType = nn.initializers.ones

  @nn.compact
  def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
    """Applies layer normalization on the input.

    It normalizes the activations of the layer for each given example in a
    batch independently, rather than across a batch like Batch Normalization.
    i.e. applies a transformation that maintains the mean activation within
    each example close to 0 and the activation standard deviation close to 1.

    Args:
      x: the inputs

    Returns:
      Normalized inputs (the same shape as inputs).

    """
    if self.hparams.quant_hparams is None:
      return nn.LayerNorm(
          epsilon=self.epsilon,
          dtype=self.dtype,
          use_bias=self.use_bias,
          use_scale=self.use_scale,
          bias_init=self.bias_init,
          scale_init=self.scale_init)(
              x)
    # To match the behavior of the upstream layernorm before the quantization
    # start step, we use float32 for intermediate computations.
    dtype = jnp.float32
    x = x.astype(dtype)

    hparams = self.hparams
    num_features = x.shape[-1]

    if self.use_scale:
      scale_param = self.param('scale', self.scale_init, (num_features,))
    if self.use_bias:
      bias_param = self.param('bias', self.bias_init, (num_features,))

    def quantized_layernorm(x):
      prec = hparams.quant_hparams.prec
      fp_quant = QuantOps.FloatQuant(is_scaled=False, fp_spec=prec)
      quant_ops = QuantOps.create_symmetric_fp(fp_quant=fp_quant, bounds=None)

      def to_quantized(x):
        return quant_ops.to_quantized(x, dtype=dtype)

      # If epsilon is too small to represent in the quantized format, we set it
      # to the minimal representative non-zero value to avoid the possibility of
      # dividing by zero.
      fp_bounds = quantization.fp_cast.get_bounds(prec.exp_min, prec.exp_max,
                                                  prec.sig_bits)
      epsilon = max(self.epsilon, fp_bounds.flush_to_zero_bound)
      quantized_epsilon = to_quantized(jnp.array(epsilon, dtype=dtype))

      # If the reciprocal of the quantized number of features is too small to
      # represent in the quantized format, we set it to the minimal
      # representative nonzero value so that the mean and variance are not
      # trivially 0.
      num_features_quantized = to_quantized(
          jnp.array(num_features, dtype=dtype))
      num_features_recip_quantized = to_quantized(
          jnp.reciprocal(num_features_quantized))
      num_features_recip_quantized = jax.lax.cond(
          jax.lax.eq(num_features_recip_quantized,
                     0.0), lambda _: quantized_epsilon,
          lambda _: num_features_recip_quantized, None)

      x_quantized = to_quantized(x)
      x_sum_quantized_reduction = quantization.quantized_sum(
          x_quantized,
          axis=-1,
          keepdims=True,
          prec=hparams.quant_hparams.reduction_prec)
      x_sum = to_quantized(x_sum_quantized_reduction)
      mean = to_quantized(x_sum * num_features_recip_quantized)
      x_minus_mean = to_quantized(x - mean)
      x_sq = to_quantized(lax.square(x_minus_mean))
      x_sq_sum_quantized_reduction = quantization.quantized_sum(
          x_sq,
          axis=-1,
          keepdims=True,
          prec=hparams.quant_hparams.reduction_prec)
      x_sq_sum = to_quantized(x_sq_sum_quantized_reduction)
      var = to_quantized(x_sq_sum * num_features_recip_quantized)
      # Prevent division by zero.
      var_plus_epsilon = to_quantized(var + quantized_epsilon)
      mul = to_quantized(lax.rsqrt(var_plus_epsilon))
      if self.use_scale:
        quantized_scale_param = to_quantized(scale_param)
        mul = to_quantized(mul * quantized_scale_param)
      y = to_quantized(x_minus_mean * mul)
      if self.use_bias:
        quantized_bias_param = to_quantized(bias_param)
        y = to_quantized(y + quantized_bias_param)
      return y.astype(self.dtype)

    # We keep  this structurally as similar as possible to
    # 'quantized_layernorm' to make it easy to compare their implementations,
    # and thus don't use convenience functions like `jnp.mean`.
    def unquantized_layernorm(x):
      num_features_recip = jnp.reciprocal(num_features)
      x_sum = jnp.sum(x, axis=-1, keepdims=True)
      mean = x_sum * num_features_recip
      x_minus_mean = x - mean
      x_sq = lax.square(x_minus_mean)
      x_sq_sum = jnp.sum(x_sq, axis=-1, keepdims=True)
      var = x_sq_sum * num_features_recip
      var_plus_epsilon = var + self.epsilon
      mul = lax.rsqrt(var_plus_epsilon)
      if self.use_scale:
        mul = mul * scale_param
      y = x_minus_mean * mul
      if self.use_bias:
        y = y + bias_param
      return y.astype(self.dtype)

    quantized_result = quantized_layernorm(x)
    unquantized_result = unquantized_layernorm(x)
    return lax.cond(self.dynamic_context.quantize_acts,
                    lambda _: quantized_result, lambda _: unquantized_result,
                    None)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements a flax module for tracking distributions and statistics."""
import math
from typing import Iterable, Optional, Union
from aqt.jax_legacy.jax import shape_utils
from aqt.jax_legacy.jax import stats
from aqt.jax_legacy.jax.utils import normalize_axes
from flax import linen as nn
import jax.numpy as jnp


def _take_subset_of_axes(x: jnp.ndarray,
                         axis: Iterable[int],
                         num_indices_per_ax: int = 10):
  """Take the first num_indices_per_ax of specified axes."""
  if not isinstance(axis, Iterable):
    axis = (axis,)
  for ax in axis:  # exclude batch dimension 0
    x = x.take(
        indices=jnp.arange(min(num_indices_per_ax, x.shape[ax])), axis=ax)
  return x


class StatsTag(nn.Module):
  """This module is intended for tracking distributions and simple statistics.

  For example, if you want track the distribution of activations and its
  statistics over the
  course of training, you can tag them with this module. You can retrieve the
  values from the state dict, and add them to tensorboard summaries for
  histogram / distribution visualization.

  Example usage:
    x = 2 * x
    stats_tag.StatsTag(update_stats=True, name='activations')(x)
    x = log(x)

  Attributes:
    update_stats: If True, will update states in this module.
    channel_axis: Axis to reduce over for computing statistics. E.g. for a
      tensor with shape (a, b, c) where c is is the channel dimension, then
      channel_axis should be 2 or -1.
    num_indices_per_ax: Number of indices to take from each channel axis.
  """
  update_stats: bool
  channel_axis: Optional[Union[int, Iterable[int]]]
  num_indices_per_ax: int = 5

  @nn.compact
  def __call__(
      self,
      x: jnp.ndarray,
      *,
      mask: Optional[jnp.ndarray],
  ) -> None:
    """Applies a tag to track distributions.

    Args:
      x: the array to compute statistics distributions over.
      mask: boolean array indicating which elements of 'x' should be included in
        the stats calculation ('True' means to include).

    Returns:
      x unchanged. The return value can also be ignored.
    """
    if mask is None:
      mask = jnp.full(x.shape, True)
    shape_utils.assert_shapes_compatible(x.shape, mask.shape)
    mask = jnp.broadcast_to(mask, x.shape)
    channel_axis = self.channel_axis
    if channel_axis is not None:
      if not isinstance(channel_axis, Iterable):
        channel_axis = (channel_axis,)
      channel_axis = normalize_axes(channel_axis, x.ndim)
      x = _take_subset_of_axes(
          x, axis=channel_axis, num_indices_per_ax=self.num_indices_per_ax)
      mask = _take_subset_of_axes(
          mask, axis=channel_axis, num_indices_per_ax=self.num_indices_per_ax)
      reduction_axis = tuple(
          [ax for ax in range(x.ndim) if ax not in channel_axis])
    else:
      reduction_axis = None

    distr_shape = ()
    if channel_axis:
      distr_shape = tuple(d for i, d in enumerate(x.shape) if i in channel_axis)

    # TODO(wanglisa): Consider adding configurability to specify which
    # statistics are collected.
    init_with_zeros = lambda shape: jnp.zeros(shape, dtype=jnp.float32)
    is_initializing = not self.has_variable('stats_tag', 'min_per_ch')
    min_per_ch = self.variable(
        'stats_tag',
        'min_per_ch',
        init_with_zeros,
        distr_shape,
    )
    max_per_ch = self.variable('stats_tag', 'max_per_ch', init_with_zeros,
                               distr_shape)
    mean_per_ch = self.variable(
        'stats_tag',
        'mean_per_ch',
        init_with_zeros,
        distr_shape,
    )
    stddev_per_ch = self.variable(
        'stats_tag',
        'stddev_per_ch',
        init_with_zeros,
        distr_shape,
    )
    absdev_per_ch = self.variable(
        'stats_tag',
        'absdev_per_ch',
        init_with_zeros,
        distr_shape,
    )
    stddev_per_ch_uncentered = self.variable(
        'stats_tag',
        'stddev_per_ch_uncentered',
        init_with_zeros,
        distr_shape,
    )
    absdev_per_ch_uncentered = self.variable(
        'stats_tag',
        'absdev_per_ch_uncentered',
        init_with_zeros,
        distr_shape,
    )
    if self.update_stats and not is_initializing:
      min_per_ch.value = jnp.min(
          jnp.where(mask, x, math.inf), axis=reduction_axis)
      max_per_ch.value = jnp.max(
          jnp.where(mask, x, -math.inf), axis=reduction_axis)
      mean_per_ch_keepdims = stats.masked_mean(
          x, mask=mask, axis=reduction_axis, paxis_name=None, keepdims=True)
      mean_per_ch.value = mean_per_ch_keepdims.squeeze(axis=reduction_axis)
      stddev_per_ch.value = jnp.sqrt(
          stats.masked_mean(
              (x - mean_per_ch_keepdims)**2,
              mask=mask,
              axis=reduction_axis,
              paxis_name=None,
              keepdims=False))
      absdev_per_ch.value = stats.masked_mean(
          jnp.abs(x - mean_per_ch_keepdims),
          mask=mask,
          axis=reduction_axis,
          paxis_name=None,
          keepdims=False)
      stddev_per_ch_uncentered.value = jnp.sqrt(
          stats.masked_mean(
              jnp.square(x),
              mask=mask,
              axis=reduction_axis,
              paxis_name=None,
              keepdims=False))
      absdev_per_ch_uncentered.value = stats.masked_mean(
          jnp.abs(x),
          mask=mask,
          axis=reduction_axis,
          paxis_name=None,
          keepdims=False)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Emulate downcasting float32/bfloat16 to custom floating-point formats."""

import dataclasses
import functools
from typing import Tuple
import jax
import jax.numpy as jnp


@dataclasses.dataclass
class FloatingPointBounds:
  """Dataclass representing the bounds for a floating-point type.

  The type is presumed to have 'flush to zero' semantics.

  Attributes:
    flush_to_zero_bound: The magnitude of the smallest representable value. If a
      logical value with an absolute value less than this is cast to this type,
      it is flushed to zero.
    saturation_bound: The magnitude of the largest representable value. If a
      logical value with an absolute value greater than this is cast to this
      type, it is clipped to this value.
  """

  flush_to_zero_bound: float
  saturation_bound: float


def get_bounds(exp_min: int, exp_max: int,
               sig_bits: int) -> FloatingPointBounds:
  """Returns the clipping bounds for a giving floating-point specification.

  Args:
    exp_min: The denormal exponent of the target format.
    exp_max: Maximum exponent of the target format (no support for infs & nans)
    sig_bits: The number of significant bits in the target format (excluding the
      hidden bit).

  Returns:
    A FloatingPointBounds dataclass.
  """
  return FloatingPointBounds(
      flush_to_zero_bound=2**exp_min,
      saturation_bound=2**exp_max * (2 - 2**-sig_bits))


@functools.partial(jax.custom_jvp, nondiff_argnums=(1, 2, 3))
def downcast_sat_ftz(
    x: jnp.ndarray,
    exp_min: int,
    exp_max: int,
    sig_bits: int,
) -> jnp.ndarray:
  """Downcast bfloat16, or float32 to a lower precision floating-point.

  - The downcast returns an argument of the *same* type as x but with
    numerical characteristics of the target fp-format.
  - The implementation (sat)urates to the largest representable number in the
    range instead of rounding to infinity.
  - Rounding mode is round-nearest-even for TPU (or configured rounding).
  - Support for special values:
    - The target fp is assumed to *not* support special values, as exp_max is
      used for the numerical range. For the sake of emulation special values are
      propagated as values on the input type, such that
      downcast(nan/inf) -> nan/inf is preserved.
  - Denormals: the target format doesn't support denormals. The minimum
    representable positive value is 2^exp_min. Denormals flush to zero (ftz).

  Args:
    x: The argument to be converted.
    exp_min: The denormal exponent of the target format.
    exp_max: Maximum exponent of the target format (no support for infs & nans)
    sig_bits: The number of significant bits in the target format (excluding the
      hidden bit).

  Returns:
   Cast of x to a lower precision float, emulating degraded precision, but of
   the same type as x.
  """
  if x.dtype not in (jnp.float32, jnp.bfloat16):
    raise ValueError('Argument is expected to be of type jnp.float32 '
                     'or jnp.bfloat16')
  # Mask for exponent bits in fp32 representation.
  exp_mask = 0x7f800000
  # NaNs / Infs have representation-value >= specials_bound.
  specials_bound = 0x7f800000
  # Binary representation of +1.0.
  one = 0x3f800000
  # Mask for mantissa bits (lower 23-bits) of fp32 representation.
  mant_mask = 0x007fffff
  xf = x.astype(jnp.float32)
  xi = xf.view(jnp.int32)
  exp = xi & exp_mask
  # Scale the argument to the unit binade.
  xi_one = (xi & mant_mask) | one
  offset = 2**(23 - sig_bits)
  # Addition of offset creates alignment to shift-off and round trailing bits.
  # Subtraction brings the rounded result back to the unit binade.
  xf_one_rnd = (xi_one.view(jnp.float32) + offset) - offset
  # Scale back to the original binade.
  xf_rnd = xf_one_rnd * exp.view(jnp.float32)
  bounds = get_bounds(exp_min=exp_min, exp_max=exp_max, sig_bits=sig_bits)
  xf_rnd_sat = jnp.minimum(xf_rnd, bounds.saturation_bound)
  # Flush denormals to zero and recover sign.
  xf_rnd_sat_ftz = jnp.sign(xf) * xf_rnd_sat * (
      xf_rnd_sat >= bounds.flush_to_zero_bound)
  xf_rnd_sat_ftz = jnp.where(exp >= specials_bound, xf, xf_rnd_sat_ftz)
  return xf_rnd_sat_ftz.astype(x.dtype)


@downcast_sat_ftz.defjvp
def _downcast_sat_ftz_jvp(
    exp_min: int, exp_max: int, sig_bits: int, primals: Tuple[jnp.ndarray],
    tangents: Tuple[jnp.ndarray]) -> Tuple[jnp.ndarray, jnp.ndarray]:
  """Computes the straight-through-estimator gradient of downcast_sat_ftz.

  This defines the approximate gradient of the `downcast_sat_ftz` function.
  Because that function is discontinuous, this defines the gradient of an
  approximate continuous version of that function. Specifically, it defines the
  gradient of:

  x = jnp.clip(x, -saturation_bound, saturation_bound)

  This ignores the rounding operations in the significand that occur in
  `downcast_sat_ftz`, and thus implements the straight-through estimator.

  Args:
    exp_min: The denormal exponent of the target format.
    exp_max: Maximum exponent of the target format (no support for infs & nans)
    sig_bits: The number of significant bits in the target format (excluding the
      hidden bit).
    primals: Primal value of the downcast_sat_ftz input. It should be a tuple
      with a single value.
    tangents: Tangent value of downcast_sat_ftz inputs evaluated at 'primals'.
      It should be a tuple with a single value.

  Returns:
    A tuple consisting of the primal output of downcast_sat_ftz and the
    corresponding tangent tensor.

  Raises:
    ValueError: `primals` or `tangents` is the wrong length.
  """
  if len(primals) != 1 or len(tangents) != 1:
    raise ValueError(
        'The primal and tangent tuples should only have one element each. '
        'This element corresponds to the one differentiable input to '
        'downcast_sat_ftz.')
  (x,), (x_dot,) = primals, tangents
  y = downcast_sat_ftz(x, exp_min=exp_min, exp_max=exp_max, sig_bits=sig_bits)

  # Differentiable approximation to downcast_sat_ftz whose gradient is used as a
  # straight-through-estimator for the downcasting operation.
  def differentiable_downcast(x):
    bounds = get_bounds(exp_min, exp_max, sig_bits)
    return jnp.clip(x, -bounds.saturation_bound, bounds.saturation_bound)

  _, y_tangent = jax.jvp(differentiable_downcast, (x,), (x_dot,))
  return y, y_tangent



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Estimation functions for compute costs of ML models."""

import contextlib
import functools
import re
from typing import Dict, Iterable, List, Optional, Tuple
from absl import flags
from aqt.jax_legacy.jax import hlo_utils
from jax._src.lax import convolution as lax_convolution
from jax._src.lax import lax
from jax.interpreters import ad
from jax.interpreters import batching
from jax.interpreters import mlir
import numpy as onp
# pylint: disable=g-direct-tensorflow-import
from tensorflow.compiler.xla.service import hlo_pb2
# pylint: enable=g-direct-tensorflow-import

FLAGS = flags.FLAGS

# in everywhere it is used in AQT once the AQT migration is done.
flags.DEFINE_bool(
    'metadata_enabled',
    default=False,
    help=('Whether to annotate quantization info in HLO metadata.'))

# We intend to use quantization information to estimate the compute cost of
# our ML model. However, currently JAX does not support transferring high-level
# information from the model to its HLO representation, e.g. via metadata.
# Therefore, using the context managers below, this info is appended to the name
# of the op in the original primitives in lax via monkey-patching. After the
# op is instantiated, the original value of the primitive is restored.
# This is not a permanent solution.


# Based on: https://jax.readthedocs.io/en/latest/jax.lax.html
class DotMetadataMonkeyPatch(contextlib.ContextDecorator):
  """Context for passing quantization data to the dot operation."""

  def __init__(self, *, lhs_prec: Optional[int], rhs_prec: Optional[int],
               rhs_is_weight: bool) -> None:
    annotation = _quantization_annotation(lhs_prec, rhs_prec, rhs_is_weight)
    self._op_name = 'dot_general_quant' + annotation

  def __enter__(self):
    # pylint: disable=protected-access
    # The following primitive accepts a name argument which is passed into
    # the HLO metadata field. Here, it is the only argument changed from
    # the original lax implementation.
    self._dot_general_p_original = lax.dot_general_p
    lax.dot_general_p = lax.standard_primitive(
        shape_rule=lax._dot_general_shape_rule,
        dtype_rule=lax._dot_general_dtype_rule,
        name=self._op_name)
    ad.defbilinear(lax.dot_general_p, lax._dot_general_transpose_lhs,
                   lax._dot_general_transpose_rhs)
    batching.primitive_batchers[lax.dot_general_p] = lax._dot_general_batch_rule
    mlir.register_lowering(lax.dot_general_p, lax._dot_general_lower)
    # pylint: enable=protected-access

  def __exit__(self, *exc):
    # Restore original primitive
    lax.dot_general_p = self._dot_general_p_original


# Based on: https://jax.readthedocs.io/en/latest/jax.lax.html
class ConvMetadataMonkeyPatch(contextlib.ContextDecorator):
  """Context for passing quantization data to the conv operation."""

  def __init__(self, *, weight_prec: Optional[int],
               act_prec: Optional[int]) -> None:
    annotation = _quantization_annotation(
        act_prec, weight_prec, rhs_is_weight=True)
    self._op_name = 'conv_general_dilated_quant' + annotation

  def __enter__(self):
    # pylint: disable=protected-access
    self._conv_general_dilated_p_original = (
        lax_convolution.conv_general_dilated_p)
    # The following primitive accepts a name argument which is passed into
    # the HLO metadata field. Here, it is the only argument changed from
    # the original lax implementation.
    lax_convolution.conv_general_dilated_p = lax.standard_primitive(
        shape_rule=lax_convolution._conv_general_dilated_shape_rule,
        dtype_rule=lax_convolution._conv_general_dilated_dtype_rule,
        name=self._op_name)
    mlir.register_lowering(lax_convolution.conv_general_dilated_p,
                           lax_convolution._conv_general_dilated_lower)
    mlir.register_lowering(
        lax_convolution.conv_general_dilated_p,
        functools.partial(lax_convolution._conv_general_dilated_lower,
                          expand_complex_convolutions=True),
        platform='cpu')
    mlir.register_lowering(
        lax_convolution.conv_general_dilated_p,
        functools.partial(lax_convolution._conv_general_dilated_lower,
                          expand_complex_convolutions=True),
        platform='gpu')
    ad.defbilinear(lax_convolution.conv_general_dilated_p,
                   lax_convolution._conv_general_dilated_transpose_lhs,
                   lax_convolution._conv_general_dilated_transpose_rhs)
    batching.primitive_batchers[lax_convolution.conv_general_dilated_p] = (
        lax_convolution._conv_general_dilated_batch_rule)
    # pylint: enable=protected-access

  def __exit__(self, *exc):
    # Restore original primitive
    lax_convolution.conv_general_dilated_p = (
        self._conv_general_dilated_p_original)


# TODO(abdolrashidi): Add support for QuantOps.FloatQuant for cost estimation.
def _quantization_annotation(lhs_prec: Optional[int], rhs_prec: Optional[int],
                             rhs_is_weight: bool) -> str:
  """Returns an annotation to be appended to the name of the quantizable op."""
  bfloat16_prec = 'bf16'

  def _replace_with_bf16_if_prec_is_none(prec: Optional[int]):
    return prec if prec is not None else bfloat16_prec

  lhs_prec = _replace_with_bf16_if_prec_is_none(lhs_prec)
  rhs_prec = _replace_with_bf16_if_prec_is_none(rhs_prec)

  quant_annotation = '_lhs{}_rhs{}_lw{}'.format(lhs_prec, rhs_prec,
                                                int(rhs_is_weight))
  return quant_annotation


def _find_lhs_shape(instr: hlo_pb2.HloInstructionProto,
                    computations: Iterable[hlo_pb2.HloComputationProto]):
  """Find the lhs shape of an instruction in HLO computations."""
  for computation in computations:
    for i in computation.instructions:
      # instr.operand_ids[0] contains the input shape.
      if i.id == instr.operand_ids[0]:
        # Input and output batch sizes must match
        assert i.shape.dimensions[0] == instr.shape.dimensions[0]
        return i.shape.dimensions
  return None


def _find_rhs_shape(instr: hlo_pb2.HloInstructionProto,
                    computations: Iterable[hlo_pb2.HloComputationProto]):
  """Find the weight shape of an instruction in HLO computations."""
  for computation in computations:
    for i in computation.instructions:
      # instr.operand_ids[1] contains the weight shape.
      if i.id == instr.operand_ids[1]:
        return i.shape.dimensions
  return None


def _estimate_weights(
    instr: hlo_pb2.HloInstructionProto,
    computations: Iterable[hlo_pb2.HloComputationProto]) -> int:
  """Estimate the number of weights in a conv or dot instruction."""
  weight_shape = _find_rhs_shape(instr, computations)
  assert weight_shape is not None
  return onp.prod(weight_shape)


def _estimate_conv_mults(
    instr: hlo_pb2.HloInstructionProto,
    computations: Iterable[hlo_pb2.HloComputationProto]) -> int:
  """Estimate the number of multiplications in a convolution instruction."""
  lhs_shape = _find_lhs_shape(instr, computations)
  assert lhs_shape is not None
  input_channels = lhs_shape[-1]
  kernel_size = onp.prod([dim.size for dim in instr.window.dimensions])
  output_conv_dims = [
      instr.shape.dimensions[i]
      for i in instr.convolution_dimension_numbers.output_spatial_dimensions
  ]
  output_image_size = onp.prod(output_conv_dims)
  output_channels = instr.shape.dimensions[
      instr.convolution_dimension_numbers.output_feature_dimension]
  return output_image_size * kernel_size * input_channels * output_channels


def _estimate_dot_mults(
    instr: hlo_pb2.HloInstructionProto,
    computations: Iterable[hlo_pb2.HloComputationProto]) -> int:
  """Estimate the number of multiplications in a dot instruction."""
  lhs_shape = _find_lhs_shape(instr, computations)
  assert lhs_shape is not None
  input_channels = lhs_shape[-1]
  output_channels = instr.shape.dimensions[-1]
  return input_channels * output_channels


def _extract_quant_info(
    instr: hlo_pb2.HloInstructionProto) -> Tuple[int, int, bool]:
  """Extracts lhs and rhs quantization precision from op metadata."""
  if instr.opcode not in _get_supported_ops():
    raise NotImplementedError('Unexpected op detected')
  # If annotated, metadata.op_type would have the following format:
  # '[original op name]_quant_w[weight prec]_a[act prec]'
  if 'quant' not in instr.metadata.op_name:
    raise NotImplementedError('Unable to parse {}'.format(
        instr.metadata.op_type))
  [(lhs_prec_str, rhs_prec_str, rhs_is_weight_str)
  ] = re.findall(r'_lhs(.*)_rhs(.*)_lw(\d+)', instr.metadata.op_name)

  def _extract_prec(prec_str):
    if prec_str.startswith('bf'):
      prec_str = prec_str[2:]
    if prec_str.isnumeric():
      prec = int(prec_str)
    else:
      raise NotImplementedError('Unable to parse {}'.format(
          instr.metadata.op_name))
    return prec

  lhs_prec = _extract_prec(lhs_prec_str)
  rhs_prec = _extract_prec(rhs_prec_str)
  rhs_is_weight = bool(int(rhs_is_weight_str))
  if lhs_prec <= 0 or rhs_prec <= 0:
    raise ValueError('HLO metadata precision annotatation must be a positive '
                     'integer.')
  return lhs_prec, rhs_prec, rhs_is_weight


def _list_supported_ops_from_hlo(
    hlo_proto: hlo_pb2.HloModuleProto) -> List[hlo_pb2.HloInstructionProto]:
  """Gather and return a list of supported quantizable ops in the HLO."""
  supported_ops = _get_supported_ops()
  target_instructions = []
  computations = hlo_proto.computations
  for computation in computations:
    for instr in computation.instructions:
      if instr.opcode in supported_ops:
        target_instructions.append(instr)
  return target_instructions


def _get_supported_ops():
  """Get the supported ops for compute and memory cost estimation."""
  # Output dictionary key is HLO instruction opcode.
  # 'estimate_instr_mult': function used to estimate the number of matrix
  # multiplications.
  # 'estimate_instr_weights': function used to estimate the number of weights.
  return {
      'convolution': {
          'estimate_instr_mult': _estimate_conv_mults,
          'estimate_instr_weights': _estimate_weights,
      },
      'dot': {
          'estimate_instr_mult': _estimate_dot_mults,
          'estimate_instr_weights': _estimate_weights,
      },
  }


def estimate_compute_cost(
    hlo_proto: hlo_pb2.HloModuleProto) -> Dict[str, float]:
  """Estimates the compute cost for the input HLO proto.

  Args:
    hlo_proto: the model's HLO representation (e.g. derived from
      hlo_utils.load_hlo_proto()) It contains all the instructions used in the
      model.

  Returns:
    a dictionary with two key-value pairs:
      'compute_cost': The sum of all compute costs for each layer
      'compute_cost_ratio_to_bfloat16': The ratio of the estimated compute cost
      to the cost in the case of no quantization (bfloat16)

  """
  # To estimate the overall compute cost of ops, we should multiply the
  # number of multiplications by bits_weights * bits_acts in each layer.

  # Gather the supported quantizable ops and their quantization parameters
  target_instructions = _list_supported_ops_from_hlo(hlo_proto)

  # Begin compute cost calculation
  compute_cost_quadratic = 0
  compute_cost_linear = 0
  bfloat16_cost_quadratic = 0  # computed as reference for comparison purposes
  bfloat16_cost_linear = 0

  supported_ops = _get_supported_ops()
  for instr in target_instructions:
    # Estimate multiplications in the op
    opcode = instr.opcode
    multiplication_count = supported_ops[opcode]['estimate_instr_mult'](
        instr, hlo_proto.computations)

    # For the model's compute cost, we use the number of multiplications used
    # in the quantizable layers used in the model, which dominate the cost
    # compared to that of other operations used in them, such as additions.
    # Also, the cost of each multiplication is proportional to the number of
    # bits in each operand, which are subject to change due to quantization.
    # Therefore, we use the weight and activation precisions in our estimation
    # as well.

    bfloat16_lhs_prec = 16
    bfloat16_rhs_prec = 16

    lhs_prec, rhs_prec, _ = _extract_quant_info(instr)
    if lhs_prec > 16:
      raise ValueError(f'Unexpected lhs precision {lhs_prec}.')
    if rhs_prec > 16:
      raise ValueError(f'Unexpected rhs precision {rhs_prec}.')

    bfloat16_cost_quadratic += multiplication_count * (
        bfloat16_lhs_prec * bfloat16_rhs_prec)
    compute_cost_quadratic += multiplication_count * lhs_prec * rhs_prec

    # Nvidia A100 only supports 16x16, 8x8 and 4x4
    # Cost of 8x8 is half of 16x16
    # Cost of 4x4 is 1/4 of 16x16
    bfloat16_cost_linear += multiplication_count * max(bfloat16_lhs_prec,
                                                       bfloat16_rhs_prec)

    prec = max(lhs_prec, rhs_prec)
    if prec <= 4:
      prec = 4
    elif prec <= 8:
      prec = 8
    elif prec <= 16:
      prec = 16

    compute_cost_linear += multiplication_count * prec
  # Return the results
  cost_ratio_to_bfloat16_quadratic = compute_cost_quadratic / bfloat16_cost_quadratic
  cost_ratio_to_bfloat16_linear = compute_cost_linear / bfloat16_cost_linear

  result_dict = {
      # quadratic compute cost. We didn't rename the keys to
      # compute_cost_quadratic to presev backwards compatibility.
      'compute_cost':
          float(compute_cost_quadratic),
      'compute_cost_ratio_to_bfloat16':
          float(cost_ratio_to_bfloat16_quadratic),
      # Linear compute cost
      'compute_cost_linear':
          float(compute_cost_linear),
      'compute_cost_ratio_to_bfloat16_linear':
          float(cost_ratio_to_bfloat16_linear),
  }

  return result_dict


def estimate_memory_cost(hlo_proto: hlo_pb2.HloModuleProto) -> Dict[str, float]:
  """Estimates the memory cost for the input HLO proto.

  Args:
    hlo_proto: the model's HLO representation (e.g. derived from
      hlo_utils.load_hlo_proto()) It contains all the instructions used in the
      model.

  Returns:
    a dictionary with two key-value pairs:
      'memory_cost': The sum of all memory costs for each layer
      'memory_cost_ratio_to_bfloat16': The ratio of the estimated memory cost to
         the cost in the case of no quantization (bfloat16)

  """
  # For the memory cost, we estimate the number of bits used for the weights,
  # which would also require the quantization precision for those weights.

  # Gather the supported quantizable ops and their quantization parameters
  target_instructions = _list_supported_ops_from_hlo(hlo_proto)

  # Begin memory cost calculation
  memory_cost = 0
  bfloat16_cost = 0  # computed as reference for comparison purposes

  supported_ops = _get_supported_ops()
  for instr in target_instructions:
    # Estimate number of weights in the op
    opcode = instr.opcode
    weight_count = supported_ops[opcode]['estimate_instr_weights'](
        instr, hlo_proto.computations)

    # Multiply number of weights by the number of bits used for each.
    bfloat16_weight_prec = 16

    _, rhs_prec, rhs_is_weight = _extract_quant_info(instr)

    if rhs_prec > 16:
      raise ValueError(f'Unexpected rhs precision {rhs_prec}.')

    if rhs_is_weight:
      memory_cost += weight_count * rhs_prec
      bfloat16_cost += weight_count * bfloat16_weight_prec

  # For dynamic matmuls, memory_cost is zero. The following check is to avoid
  # division by zero. We set the ratio for 0/0 to 1.
  if bfloat16_cost == 0 and memory_cost == 0:
    cost_ratio_to_bfloat16 = 1
  else:
    cost_ratio_to_bfloat16 = memory_cost / bfloat16_cost

  result_dict = {
      'memory_cost': float(memory_cost),
      'memory_cost_ratio_to_bfloat16': float(cost_ratio_to_bfloat16)
  }
  return result_dict


def estimate_costs_of_dot_and_conv_ops_from_jax_fn(
    fn, *fn_args, **fn_kwargs) -> Dict[str, float]:
  """Wrapper function around estimate_compute_and_memory_cost().

  Will generate hlo proto from jax function, and call
  estimate_compute_and_memory_cost() on it.

  Args:
    fn: the function for which the HLO is to be produced.
    *fn_args: the function's args.
    **fn_kwargs: the function's kwargs.

  Returns:
    A dictionary with compute cost and memory cost data.
    See estimate_compute_and_memory_cost() docstring for details.
  """
  FLAGS.metadata_enabled = True
  hlo_module_proto = hlo_utils.load_hlo_proto_from_jax_fn(
      fn, *fn_args, **fn_kwargs)
  cost_dict = estimate_compute_cost(hlo_module_proto)
  memory_cost_dict = estimate_memory_cost(hlo_module_proto)
  cost_dict.update(memory_cost_dict)
  FLAGS.metadata_enabled = False
  return cost_dict



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Resnet50 weights only quantized to 4 bits model."""

from aqt.jax_legacy.jax.imagenet.configs import base_config


def get_config(quant_target=base_config.QuantTarget.WEIGHTS_ONLY):
  config = base_config.get_config(
      imagenet_type=base_config.ImagenetType.RESNET50,
      quant_target=quant_target)
  config.weight_prec = 4
  config.model_hparams.conv_init.weight_prec = 8
  config.model_hparams.dense_layer.weight_prec = 8
  return config
